
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cbp.h"

#define M_PI2	(M_PI*M_PI)
#define delta2	(delta*delta)
#define k2	((double)k*(double)k)

double	Ramachandran(delta,k)
double	delta;
int	k;
{
	if (k==0)
	    return 1.0/(4.0*delta2);
	else if (k&1)
	    return -1.0/(M_PI2*delta2*k2);
	else
	    return 0.0;
}

double	Shepp(delta,k)
double	delta;
int	k;
{
	return 2.0/(M_PI2*delta2*(1.0-4.0*k2));
}

double	Chesler(delta,k)
double	delta;
int	k;
{
	if (k==0)
	    return (1.0/4.0-1.0/M_PI2)/(2.0*delta2);
	else if (k==1 || k==(-1))
	    return (1.0/8.0-1.0/M_PI2)/(2.0*delta2);
	else if (k&1)
	    return -1.0/(2.0*M_PI2*delta2*k2);
	else
	    return -(k2+1.0)/(2.0*M_PI2*delta2*(k2-1.0)*(k2-1.0));
}

#ifndef	SIMPLE_CONVOLUTION
static void	FFT(N,Z,Exp,sign)
int		N;
Complex		*Z,*Exp;
double		sign;
{
	int	i,j,k,l,m,n,p,q,r,N1=N-1,N2=N>>1;
	Complex	c;
	double	Cos,Sin;

	for (i=1; i<N1; i++) {
	    j=0; for (k=N2, l=1; l<=i; l<<=1, k>>=1) if (i&l) j|=k;
	    if (i<j) { c=Z[i]; Z[i]=Z[j]; Z[j]=c; }
	}
	for (i=2, j=1, k=N2, l=N; l>1; l=k, k>>=1, j=i, i<<=1)
	    for (m=n=0; n<j; n++, m+=l) {
		Cos=     Exp[m].re;
		Sin=sign*Exp[m].im;
		for (p=n, q=0; q<k; q++, p+=i) {
		    r=p+j; c.re=Z[r].re*Cos-Z[r].im*Sin;
			   c.im=Z[r].re*Sin+Z[r].im*Cos;
			   Z[r].re=Z[p].re-c.re;
			   Z[r].im=Z[p].im-c.im;
			   Z[p].re+=c.re;
			   Z[p].im+=c.im;
		}
	    }
}
#endif

#define ALLOC(noe,ptr)	malloc((noe)*sizeof(*(ptr)))

CBP	*SetUpCBP(N,M,delta,r0,theta0,ifilter)
//CBP	*SetUpCBP(N,M,delta,r0,theta0)
int	N,M;
double	delta,r0,theta0;
{
	int	m,n;
	CBP	*cbp;
	double	theta,R,Filter();

#ifndef	SIMPLE_CONVOLUTION
	int	L,L2;
	double	phase;

	for (L=2; L<N; L<<=1) ; L2=L<<1;
#else
	int	N21=N*2-1,N1=N-1;
#endif

	if ((cbp     =(CBP     *)ALLOC(1  ,cbp     ))==NULL ||
	    (cbp->pq =(Float  **)ALLOC(M  ,cbp->pq ))==NULL ||
#ifndef	SIMPLE_CONVOLUTION
	    (cbp->G  =(Complex *)ALLOC(L2 ,cbp->G  ))==NULL ||
	    (cbp->Z  =(Complex *)ALLOC(L2 ,cbp->Z  ))==NULL ||
	    (cbp->Exp=(Complex *)ALLOC(L2 ,cbp->Exp))==NULL ||
#else
	    (cbp->g  =(double  *)ALLOC(N21,cbp->g  ))==NULL ||
	    (cbp->z  =(double  *)ALLOC(N,  cbp->z  ))==NULL ||
#endif
	    (cbp->Cos=(double  *)ALLOC(M  ,cbp->Cos))==NULL ||
	    (cbp->Sin=(double  *)ALLOC(M  ,cbp->Sin))==NULL) return NULL;

#ifdef	CONCENTRIC_SYMMETRY
	if ((cbp->pq[0]=(Float *)ALLOC(N+1,cbp->pq[0]))==NULL) return NULL;

	for (m=1; m<M; m++) cbp->pq[m]=cbp->pq[0];

	for (n=0; n<=N; n++) cbp->pq[0][n]=0.0;
#else
	for (m=0; m<M; m++) {
	    if ((cbp->pq[m]=(Float *)ALLOC(N+1,cbp->pq[m]))==NULL) return NULL;

	    for (n=0; n<=N; n++) cbp->pq[m][n]=0.0;
	}
#endif
	cbp->N=N;
#ifndef	SIMPLE_CONVOLUTION
	cbp->L=L;
#endif
	cbp->delta=delta;
	cbp->r0delta=r0/delta;
	R=r0+delta*(double)(N-1); if (R>(-r0)) R=(-r0); cbp->R2=R*R;

#ifndef	SIMPLE_CONVOLUTION
	for (n=0; n<L2; n++) {
	    if (ifilter == 1) cbp->G[n].re = Chesler(delta,n-L);
	    if (ifilter == 2) cbp->G[n].re = Shepp(delta,n-L);
	    if (ifilter == 3) cbp->G[n].re = Ramachandran(delta,n-L);
//	    cbp->G[n].re=Filter(delta,n-L);
	    cbp->G[n].im=0.0;

	    phase=M_PI*(double)n/(double)L2; cbp->Exp[n].re=cos(phase);
					     cbp->Exp[n].im=sin(phase);
	}
	FFT(L2,cbp->G,cbp->Exp,-1.0);
#else
	if (ifilter == 1) for (n=(-N1); n<=N1; n++) cbp->g[n+N1]=Chesler(delta,n);
	if (ifilter == 2) for (n=(-N1); n<=N1; n++) cbp->g[n+N1]=Shepp(delta,n);
	if (ifilter == 3) for (n=(-N1); n<=N1; n++) cbp->g[n+N1]=Ramachandran(delta,n);
//	for (n=(-N1); n<=N1; n++) cbp->g[n+N1]=Filter(delta,n);
#endif
	cbp->M=M;
	cbp->theta0=theta0;
	cbp->M_PIM=M_PI/(double)M;

	for (m=0; m<M; m++) {
	    theta=theta0+cbp->M_PIM*(double)m; cbp->Cos[m]=cos(theta);
					       cbp->Sin[m]=sin(theta);
	}
	return cbp;
}

void	Convolution(cbp)
CBP	*cbp;
{
	int	M,m,n,N=cbp->N;
	Float	**pq,*p,*q;

#ifndef	SIMPLE_CONVOLUTION
	int	L=cbp->L,L2=L<<1;
	Complex	*Z,*G,c,*cbpZ=cbp->Z,*cbpG=cbp->G,*cbpExp=cbp->Exp;
	double	deltaL2=cbp->delta/(double)L2;
#else
	int	j;
	double	*zj,*gj,z,*g,
		*cbpz=cbp->z,*cbpg=(&(cbp->g[N-1])),delta=cbp->delta;
#endif

#ifdef	CONCENTRIC_SYMMETRY
	M=1;
#else
	M=cbp->M;
#endif

	for (pq=cbp->pq, m=0; m<M; m++, pq++) {
#ifndef	SIMPLE_CONVOLUTION
	    for (Z=cbpZ,  n=0; n<L; n++,      Z++)   Z->re=      Z->im=0.0;
	    for (p=(*pq), n=0; n<N; n++, p++, Z++) { Z->re=(*p); Z->im=0.0; }
	    for (            ; n<L; n++,      Z++)   Z->re=      Z->im=0.0;

	    FFT(L2,cbpZ,cbpExp,-1.0);

	    for (Z=cbpZ, G=cbpG, n=0; n<L2; n++, G++, Z++) {
		c.re=G->re*Z->re-G->im*Z->im;
		c.im=G->re*Z->im+G->im*Z->re; *Z=c;
	    }
	    FFT(L2,cbpZ,cbpExp, 1.0);

	    for (q=(*pq), Z=cbpZ, n=0; n<N; n++, Z++, q++) *q=deltaL2*Z->re;
#else
	    for (zj=cbpz, gj=cbpg, j=0; j<N; j++, gj++, zj++) {
		z=0.0;
		for (p=(*pq), g=gj, n=0; n<N; n++, g--, p++) z+=(*g*(*p));
		*zj=delta*z;
	    }
	    for (q=(*pq), zj=cbpz, j=0; j<N; j++, zj++, q++) *q=(*zj);
#endif
	}
}

double	BackProjection(cbp,x,y)
CBP	*cbp;
double	x,y;
{
	double	r0,*Cos,*Sin,r,f=0.0;
	int	M,m,n;
	Float	**pq,*q;

	if (x*x+y*y<=cbp->R2) {
	    x/=cbp->delta;
	    y/=cbp->delta; r0=cbp->r0delta;
	    M=cbp->M; pq=cbp->pq; Cos=cbp->Cos; Sin=cbp->Sin;
	    for (m=0; m<M; m++, Sin++, Cos++, pq++) {
		r=x*(*Cos)+y*(*Sin)-r0; n=(int)r; q=(*pq);
		f+=((r-(double)n)*(q[n+1]-q[n])+q[n]);
	    }
	}
	return f*cbp->M_PIM;
}

Float	**BPImage(cbp)
CBP	*cbp;
{
	Float	**f,**fv,*fh,*q;
	int	*h0,*Nh,v,h,m,n,N=cbp->N,v1=N-1,v2=0,M=cbp->M;
	double	*xy,r,Rv2,Rv,Cos,Sin,N12=(double)(N-1)/2.0,
				     R2=cbp->R2/(cbp->delta*cbp->delta),
				     r0=cbp->r0delta,
				     M_PIM=cbp->M_PIM;

	if ((f =(Float **)ALLOC(N,f ))==NULL ||
	    (xy=(double *)ALLOC(N,xy))==NULL ||
	    (h0=(int    *)ALLOC(N,h0))==NULL ||
	    (Nh=(int    *)ALLOC(N,Nh))==NULL) return NULL;

	for (fv=f, v=0; v<N; v++, fv++) {
	    if ((*fv=(Float *)ALLOC(N,*fv))==NULL) return NULL;

	    for (fh=(*fv), h=0; h<N; h++, fh++) *fh=0.0;

	    r=xy[v]=(double)v-N12;
	    if ((Rv2=R2-r*r)>=0.0) {
		if (v<v1) v1=v;
		if (v>v2) v2=v;

		Rv=sqrt(Rv2); Nh[v]=(int)(N12+Rv)-(h0[v]=(int)ceil(N12-Rv));
	    }
	}
	for (m=0; m<M; m++) {
//		if(m<1332||m>1479){
//	for (m=0; m<(int)(M*169./180.); m++) {
	    q=cbp->pq[m]; Cos=cbp->Cos[m]; Sin=cbp->Sin[m];
	    for (v=v1, fv=f+v; v<=v2; v++, fv++) {
			h=h0[v]; fh=(*fv)+h; r=xy[h]*Cos-xy[v]*Sin-r0;
			for (h=Nh[v]; h>=0; h--, fh++, r+=Cos) {
		    	n=(int)r; *fh+=((r-(double)n)*(q[n+1]-q[n])+q[n]);
			}
	    }
//		}
	}
	for (v=v1, fv=f+v; v<=v2; v++, fv++)
	    for (fh=(*fv)+h0[v], h=Nh[v]; h>=0; h--, fh++) *fh*=M_PIM;

	free(Nh); free(h0); free(xy); return f;
}

int	ReGridingCP(cbp)
CBP	*cbp;
{
	int	N,M,segments,m,n,n0,s;
	Float	**pq,**new1,*old1,*old2,*new2;
	double	r0,ratio,theta,NM_PI2=(double)cbp->N*M_PI/2.0;

	if ((double)cbp->M>=NM_PI2) return 1;

	N=cbp->N;
	M=cbp->M; segments=(int)ceil(NM_PI2/(double)M); cbp->M=M*segments;

	if ((pq   =(Float **)ALLOC(M+1,pq   ))==NULL ||
	    (pq[M]=(Float  *)ALLOC(N+1,pq[M]))==NULL) return 0;

	for (m=0; m<M; m++) pq[m]=cbp->pq[m];

	for (n=0; n<=N; n++) {
	    r0=(-2.0*cbp->r0delta-(double)n);
	    if (r0>=0.0 && r0<=(double)(N-1)) {
		n0=(int)floor(r0);
		pq[M][n]=pq[0][n0]+(r0-(double)n0)*(pq[0][n0+1]-pq[0][n0]);
	    }
	    else
		pq[M][n]=0.0;
	}

	free(cbp->pq);
	if ((cbp->pq=new1=(Float **)ALLOC(cbp->M,cbp->pq))==NULL) return 0;

	for (old1=pq[0], m=1; m<=M; m++, old1=old2) {
	    *new1=old1; old2=pq[m];
	    for (++new1, s=1; s<segments; s++, new1++) {
		if ((*new1=new2=(Float *)ALLOC(N+1,new2))==NULL) return 0;

		ratio=(double)s/(double)segments;
		for (n=0; n<=N; n++) *(new2++)=old1[n]+ratio*(old2[n]-old1[n]);
	    }
	}
	free(pq[M]); free(pq);

	free(cbp->Cos);
	free(cbp->Sin);
	if ((cbp->Cos=(double *)ALLOC(cbp->M,cbp->Cos))==NULL ||
	    (cbp->Sin=(double *)ALLOC(cbp->M,cbp->Sin))==NULL) return 0;

	cbp->M_PIM=M_PI/(double)cbp->M;

	for (m=0; m<cbp->M; m++) {
	    theta=cbp->theta0+cbp->M_PIM*(double)m; cbp->Cos[m]=cos(theta);
						    cbp->Sin[m]=sin(theta);
	}
	return 1;
}

void	ClearCBP(cbp,f)
CBP	*cbp;
Float	**f;
{
	int	n,m,N=cbp->N,M=cbp->M;
	Float	**pq=cbp->pq;

	if (f!=NULL) {
	    for (n=0; n<N; n++) free(f[n]); free(f);
	}

	free(cbp->Sin); free(cbp->Cos);

#ifndef	SIMPLE_CONVOLUTION
	free(cbp->Exp); free(cbp->Z); free(cbp->G);
#else
	free(cbp->z); free(cbp->g);
#endif

#ifdef	CONCENTRIC_SYMMETRY
{
	int	m0;

	for (m=0; m<M; m++) {
	    for (m0=0; m0<m; m0++) if (pq[m]==pq[m0]) break;

	    if (m==m0) free(pq[m]);
	}
}
#else
	for (m=0; m<M; m++) free(pq[m]);
#endif
	free(pq); free(cbp);
}
