// Convodution back projection method by Tsukasa Nakano, GSJ
// Required files are sinogram data, sino.tmp.

/*----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "cbp.h"
#include <malloc.h>
 
/*----------------------------------------------------------------------*/
#define INTEL

/*----------------------------------------------------------------------*/

// main data for read transmitted images (data[y][x])
//short	data[MAXPIXL][MAXPIXL];
double	*prj_temp;

// output file name
char	fn[15];

// region of calculation in raw data(transmitted images)
short	x1, x2, z0, z1;

// filter number (Chesler, Shepp, Ramachandran)
int		ifilter;

// common parameter
int		NST;
double	r0;

// total absorption
double	TAM, TASD;


/*----------------------------------------------------------------------*/

static int StoreProjection(cbp)
CBP			*cbp;
{
	int		M     = cbp->M,
			N     = cbp->N;
	double	*Cos  = cbp->Cos,
			*Sin  = cbp->Sin,
			delta = cbp->delta;
	double	**p   = cbp->pq;

	long	i, j;

// initialize p
	for(j=0; j<M; j++) for(i=0; i<N; i++) p[j][i]=0.0;

// p ɑ
	for (j=0; j<M; j++){
//	for (j=150; j<M-150; j++){
		for (i=0; i<N; i++){
			if(abs(*(prj_temp+N*j+i))<100){
			p[j][i]=*(prj_temp+N*j+i);
			}
		}
//		}
//		printf("%d\t%d\t%lf\t%lf\n",500 ,j, *(prj_temp+(N*j+(int)((1-N)/2-r0/delta)+500)), p[j][500]);
	}
	free(prj_temp);
	return (0);
	exit(0);
}


/*----------------------------------------------------------------------*/

static void Error(msg)
char        *msg;
	{
		fputs(msg,stderr);
		fputc('\n',stderr);
		exit(1);
	}

/*----------------------------------------------------------------------*/

#ifndef CLOCKS_PER_SEC
#define CLOCKS_PER_SEC	1000000
extern long clock();
#endif

#define CLOCK()		((double)clock()/(double)CLOCKS_PER_SEC)
#define DESC_LEN	256

int main(argc,argv)
int		argc;
char	**argv;
{
	CBP			*cbp;
	short		N;				// reconstructed image full size (N*N)
	double		N12;			// N12 = (N-1)/2
	short		ln;				// layer
	short		M;				// number of projection

	double		Clock;			// timer setting
	double		**f;			// full size of reconstructed image

	int			vv, hh;			// 
	double		delta, theta0;
	double		*rec_temp;
	int			i_tmp;			// temporary parameter for integer
	FILE		*fp;
	char		line[20], *ss;
	long i;

	if (argc>2)
		Error(" usage : ct_cbp\n");

//open input files
	if((fp = fopen("sino.tmp","r")) == NULL){ 
		printf("can not open %s for input\n", fn);
		exit(1);
	}
	ss=fgets(line, 20, fp); if (i_tmp=strlen(line)>1){sscanf(line, "%d", &N);}
	ss=fgets(line, 20, fp); if (i_tmp=strlen(line)>1){sscanf(line, "%d", &M);}
	ss=fgets(line, 20, fp); if (i_tmp=strlen(line)>1){sscanf(line, "%d", &ln);}
	ss=fgets(line, 20, fp); if (i_tmp=strlen(line)>1){sscanf(line, "%lf", &r0);}
	ss=fgets(line, 20, fp); if (i_tmp=strlen(line)>1){sscanf(line, "%lf", &TAM);}
	ss=fgets(line, 20, fp); if (i_tmp=strlen(line)>1){sscanf(line, "%lf", &TASD);}
	fclose(fp);

	sprintf(fn,"s%04d.sin", ln);

	printf("calclation layer\t: %d\n", ln);
	printf("sinogram file name\t: %s\n", fn);
	printf("Width of image\t\t: %d\n", N);
	printf("Number of projection\t: %d\n", M);
	printf("Rotation center\t\t: %f\n", (float)r0);

//open input files
	prj_temp = (double *)malloc(N*M*sizeof(double));
	if((fp = fopen(fn,"rb")) == NULL){
		printf("can not open %s for input\n", fn);
		return(-1);
	}
	if ((i = fread(prj_temp , sizeof(double), N*M, fp))!=N*M){
		printf("error in reading sinogram (%d)\n", (int)i);
		return(-2);
	}
	fclose(fp);


	r0=-1*r0;
	delta = 1.0; //normilized data

	N12		= (double)(N-1) / 2.0;
	if (argc==2){
		theta0 = atof(argv[1]);
	} else{
		theta0 = 0.0;
	}
	ifilter	= 1; // Chesler
//	ifilter	= 2; // Shepp-Logan
//	ifilter	= 3; // Ramachandran

//	Clock=CLOCK();
	if ((cbp=SetUpCBP(N, M, delta, r0, theta0*M_PI/180.0, ifilter))==NULL)
		Error("memory allocation error for CBP.");
//	fprintf(stderr,"SetUpCBP\t\t%lf / sec\n",CLOCK()-Clock);

	cbp->r0delta = r0 / delta;

//	Clock=CLOCK();
	i_tmp=StoreProjection(cbp);
//	fprintf(stderr,"Store Projection \t%lf / sec\n",CLOCK()-Clock);

// Convolution ̏v
	Clock=CLOCK();
	Convolution(cbp);
	fprintf(stderr,"Convolution\t\t: %lf / sec\n",CLOCK()-Clock);

//	if (!ReGridingCP(cbp)) Error("ReGridingCP() memory allocation error.");

	rec_temp= (double *)malloc(N*N*sizeof(double));

#define BPIMAGE

	Clock=CLOCK();
#ifdef	BPIMAGE
	{
	if ((f=BPImage(cbp))==NULL)
	    Error("memory allocation error for BPImage.");

	for (vv=0; vv<N; vv++){
		for (hh=0; hh<N; hh++){
			*(rec_temp+N*vv+hh) = f[vv][hh];
		}
	}
}
#else	/* ifndef BPIMAGE */
{
	if ((f=(Float **)malloc(N*sizeof(*f)))==NULL)
	    Error("memory allocation error for BackProjection (1).");

	for (vv=0; vv<N; vv++)
	    if ((f[vv]=(Float *)malloc(N*sizeof(**f)))==NULL)
		Error("memory allocation error for BackProjection (2).");

	for (vv=0; vv<N; vv++){
		for (hh=0; hh<N; hh++){
			f[vv][hh] = BackProjection(cbp, ((double)hh-N12)*delta, (N12-(double)vv)*delta);
//			if (f[vv][hh]>f_MAX) f_MAX = f[vv][hh];
//			if (f[vv][hh]<f_min) f_min = f[vv][hh];
			*(rec_temp+N*vv+hh) = f[vv][hh];
		}
	}
}
#endif	/* BPIMAGE */
	fprintf(stderr,"BackProjection\t\t: %lf / sec\n",CLOCK()-Clock);

// output reconstructed data
	sprintf(fn,"r%04d.rec", ln);
	if((fp = fopen(fn,"wb")) == NULL){ 
		printf("can not open %s for output\n", fn);
		return(-3);
	}

	hh=0;
	for (vv=0;vv<N;vv++){
		for (hh=0;hh<N;hh++){
			if((i_tmp = fwrite((rec_temp+N*vv+hh) , sizeof(double), 1, fp))!=1){
				printf("error writing rec data (%d)", i_tmp);
				return (-4);
			}
		}
	}

	fclose(fp);

// output the detail of calculation

	if((fp = fopen("rec.tmp","w")) == NULL){ 
		printf("can not open rec.tmp for output\n");
		return(-4);
	}
	fprintf(fp, "%d\n", ln);
	fprintf(fp, "%d\n", N);
	fprintf(fp, "%d\n", M);
	fprintf(fp, "%d\n", N);
	fprintf(fp, "%lf\n", r0);
	fprintf(fp, "%d\n", ifilter);
//	fprintf(fp, "%d\n", 1); // HAN tB^[cheslar^CvɌŒB
	fprintf(fp, "%lf\n", TAM); // Mean of Total Absorption from sinogram
	fprintf(fp, "%lf\n", TASD); // SD of Total Absorption from sinogram
	fclose(fp);

// free memory
	ClearCBP(cbp,f);

// remove winogram file

	sprintf(fn,"s%04d.sin", ln);
	if(remove( fn ) == -1 ){
      printf( "%s could not be deleted.",fn );
	}

	exit (0);
}
